package com.xefer.rebel;

import java.io.*;
import java.lang.*;
import java.util.*;

import javax.servlet.http.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import org.apache.xpath.*;
import org.w3c.dom.*;
 
/**
 * Constructs XBELElements
 *
 */
 
class XBELElementFactory {

  /**
   *
   * newElement
   *
   * Convert the given url path into an XPath then extract the appropriate node
   * from the given XML Document
   *
   * @param istream <code>InputStream</code> an XML document containing resource elements
   * @return <code>XBELElement</code> the Element (if any) referenced by the converted path
   *
   */

   static public XBELElement newElement(InputStream istream) throws ResourceException {
  
    Document document;
    
    try {
    
      // parse the supplied input stream as an XML document
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      
      factory.setExpandEntityReferences(false);
      factory.setValidating(true);
      
      DocumentBuilder builder = factory.newDocumentBuilder();
      
      builder.setErrorHandler(null);
      
      document = builder.parse(istream);
    
    } catch (ParserConfigurationException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Can't build parser for input document");
    
    } catch (org.xml.sax.SAXException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Can't parse supplied document");
    
    } catch (IOException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Can't open supplied document");
    
    } 
    
    return newElement(document.getDocumentElement());
    
  }

  /**
   *
   * newElement
   *
   * Convert the given url path into an XPath and extract the appropriate node
   * from the given XML Document
   *
   * @param document <code>Document</code> a url encoded path to be converted to an xpath 
   * @param path  <code>String</code> a url encoded path to be converted to an xpath 
   * @return <code>XBELElement</code> the Element (if any) referenced by the converted path
   *
   */
   
  static public XBELElement newElement(Document document, String path) throws ResourceException {
  
    Element resource;
    
    try {
    
      resource = (Element) XPathAPI.selectSingleNode(document, convertURLtoXPath(path));

      if(resource == null) {
        throw new ResourceException(HttpServletResponse.SC_NOT_FOUND, "Unknown Resource");
      }
    
    } catch (TransformerException ex) {
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Unable to Transform");
    }
    
    return newElement(resource);
    
  }
  
  /**
   *
   * getElement
   *
   * Convert the given url path into an XPath and extract the appropriate node
   * from the given XML Document.  Returns null if no resource found at the given
   * path
   *
   * @param document <code>Document</code> a url encoded path to be converted to an xpath 
   * @param path  <code>String</code> a url encoded path to be converted to an xpath 
   * @return <code>XBELElement</code> the Element (if any) referenced by the converted path
   *
   */
   
  static public XBELElement getElement(Document document, String path) throws ResourceException {
  
    Element resource;
    
    try {
    
      resource = (Element) XPathAPI.selectSingleNode(document, convertURLtoXPath(path));
    
    } catch (TransformerException ex) {
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Unable to Transform");
    }

    if(resource != null) {
      return newElement(resource);
    }
    
    return null;    
    
  }
   
   
  /**
   *
   * newElement
   *
   * Convert the given url path into an XPath and extract the appropriate node
   * from the given XML Document
   *
   * @param document <code>Document</code> a url encoded path to be converted to an xpath 
   * @return <code>XBELElement</code> the Element (if any) referenced by the converted path
   *
   */

  static public XBELElement newElement(Document document) throws ResourceException {
  
    return newElement(document.getDocumentElement());
    
  }
   
  /**
   *
   * newElement
   *
   * Convert the given url path into an XPath and extract the appropriate node
   * from the given XML Document
   *
   * @param element <code>Element</code> a url encoded path to be converted to an xpath 
   * @return <code>XBELElement</code> the Element (if any) referenced by the converted path
   *
   */
  
  static public XBELElement newElement(Element element) throws ResourceException {
  
    String tag = element.getTagName();
    
    if(tag == "bookmark") {    
      return new XBELBookmark(element);    
    }
    
    if(tag == "folder") {   
      return new XBELFolder(element);   
    }
    
    if(tag == "xbel") {
      return new XBELDocument(element);
    }
    
    // todo: have generic element for schema extensions
    //throw new ResourceException(HttpServletResponse.SC_BAD_REQUEST, "Invalid schema");
    
    return new XBELGenericElement(element);
  }
  
  /**
   *
   * generateElement
   *
   * Construct an XBELElement at the given path
   *
   * @param document <code>Document</code> the owning document.
   * @param path  <code>String</code> a url where the resource is to be created
   * @return <code>XBELElement</code> the Element generated
   *
   */
   
  static public XBELElement generateElement(Document document, String path) throws ResourceException {
  
    XBELElement resource;
    XBELElement found;
    XBELElement entity;
    XBELElement holder;
    XBELElement temporary;

    String workpath = new String();
    String goodpath = new String();
    String testpath = null;
    String name     = null;

    int index = 0;
    
    // This is a bit complicated because of all the string manipulation
    // and the checking for and building of resources. It could perhaps be made
    // clearer but the general algorithm is this: read the
    // given path one /element/ at a time and check if the resource
    // exists. If it does, continue going down the path until
    // encountering one that does not exist. Once reaching an
    // /element/ that doesn't exist, begin constructing a tree
    // of folders until either a final folder or bookmark is reached.
    // At that point, attach the first new element as a child of the last old element,
    // then return the last element created.  That element will be overwritten
    // by whatever entity has been PUT to the URL. At that point, opreration
    // continues as if it were a PUT to any existing resource.
    
    XBELDocument xbel = new XBELDocument(document);
    
    // we're putting to the root document so first check to see if this
    // is a new document altogether
    if((path == null) | path.equals("/")) {
      return xbel;
    }
    
    // just make sure we don't have an existing resource
    // already and skip all the hassle
    resource = getElement(document, path);
    
    if(resource != null) {
      return resource;
    }

    // work through the current path to see where we need to 
    // start adding elements, always maintaining the current
    // "good path" of known folders

    StringTokenizer tokens = new StringTokenizer(path, "/", true);

    found = null;
    
    do {
    
      String slash = tokens.nextToken(); // the slash
      String rname = tokens.nextToken(); // the name
      
      // keep building the workpath, creating the testpath
      // here checks to see if there are more tokens (which by
      // definition must be a '/') and adds it without actually
      // getting it.
      workpath = workpath + slash + rname;
      testpath = workpath + (tokens.hasMoreTokens() ? "/" : "");
      
      resource = getElement(document, testpath);
      
      if(resource != null) {
        goodpath = workpath; // not null, so this path is good
        found    = resource; // hang onto the deepest good folder
      }
    
    } while (resource != null);

    // remove the known good path and work on building resource 
    // for the unknown, new resources
    
    workpath = path.substring(goodpath.length()); 

    System.out.println("[Put] New Workpath: " + workpath);

    boolean created = false;
    holder          = null;
    temporary       = null;
    entity          = null;
    
    tokens = new StringTokenizer(workpath, "/", true);
    
    while(tokens.hasMoreTokens()) {
    
      String bslash = tokens.nextToken(); // the slash
      if(tokens.hasMoreTokens() == false) { 
        break;                            // if that was the last trailing slash, we're done
      }
      String rname  = tokens.nextToken(); // the name
      
      if(tokens.hasMoreTokens()) {        // if no more, then must be bookmark
        entity = new XBELFolder(document);  
        entity.setId(rname);      
      }
      else {
        entity = new XBELBookmark(document); 
        entity.setId(rname);      
      }
      
      // hang onto the first entity created; that's the one to be added to the last folder found
      if(created == false) {
        holder    = entity;
        temporary = entity;
        created   = true;
      }
      else {
        temporary.addChild(entity);
        temporary = entity;
      }
      
    }
    
    if(found == null) {
      xbel.addChildDirect(holder);
    }
    else {
      found.addChildDirect(holder);
    }
    
    // return the new resource we're going to put to
    // which is the last one created
    return entity;

  }
   
  
  /**
   *
   * Convert the given url path into an XPath 
   *
   * @param path  <code>String</code> a url encoded path to be converted to an xpath 
   * @return <code>String</code> an XPath representation of the String
   *
   */

  static public String convertURLtoXPath(String path) {

    String xpath = new String("//xbel");

    if(path == null) {
      return xpath;
    }
    
    StringTokenizer tokenizer = new StringTokenizer(path, "/", true);

    tokenizer.nextToken(); // pinch off the leading "/"

    while(tokenizer.hasMoreTokens()) {

      String tokenOne;
      String tokenTwo;

      tokenOne = tokenizer.nextToken();

      if(tokenizer.hasMoreTokens()) {
        tokenTwo = tokenizer.nextToken(); // pinch off the "/"
        xpath += "/folder[@id='" + tokenOne + "']";
      }
      else {
        xpath += "/bookmark[@id='" + tokenOne + "']";
      }

    }

    return xpath;

  }

}